----------------------------------------------------------------------------------
-- Company:        CERN, BE-CO-HT
-- Engineer:       Carlos Gil Soriano
--
-- Create Date:    00:47:34 10/26/2011
-- Design Name:    i2c bit recognition fsm
-- Module Name:    i2c_bit - behav
-- Project Name:   Level Conversion Circuits
-- Target Devices: Spartan 6
-- Tool versions:
-- Description:    This is the fsm for the recognition of a bit. It offers some
--                 interrupting lines to the upper level (i2c_slave_core.vhd)
--                 which is responsible for managing byte level transactions
--
-- Dependencies:   The inputs should be debounced: i2c_debouncer.vhd
--
-- Revision:
-- Revision 0.01 - File Created
--          0.1  - Module works with no reported issues.
--          1.0  - Code revamped. Now is clearer and easier to read.
-- Additional Comments:
--
----------------------------------------------------------------------------------
library ieee;

use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity i2c_bit is
  port 
  (
    rst_i  : in   std_logic;
    clk_i  : in   std_logic;

    sda_i     : in   std_logic;
    sda_o     : out std_logic;
    sda_en_o  : out std_logic;
    scl_i     : in   std_logic;
    scl_o     : out std_logic;
    scl_en_o  : out std_logic;

    tx_i      : in   std_logic;
    rx_o      : out  std_logic;

    start_o   : out  std_logic;
    stop_o    : out  std_logic;
    done_o    : out  std_logic
  );
end entity i2c_bit;

architecture behav of i2c_bit is

  component i2c_debouncer is
    generic 
    (
      g_LENGTH       : NATURAL := 6
    );
    port 
    (
      rst         : in  std_logic;
      clk         : in  std_logic;
      input       : in  std_logic;
      output      : out std_logic;
      glitch_mask : in  std_logic_VECTOR(g_LENGTH - 1 downto 0)
    );
  end component i2c_debouncer;


   type t_state is (
                     R0_RESET,
                     S0_IDLE,
                     S1A_HIGH_TMP,
                     S1A_HIGH,
                     S1B_LOW_TMP,
                     S1B_LOW,
                     S2A_START_TMP,
                     S2A_START,
                     S2B_STOP_DETECT,
                     Q1_ERROR
                   );

   -- It specifies the maximum number of stages that will be employed for
   -- deglitching. Clocked with clk_i
   constant c_MAX_GLITCH_DELAY : NATURAL := 6;
   
   -- Three delay stages out of six
   constant c_GLITCH_MASK      : std_logic_VECTOR (5 downto 0) := "000111";

   signal sda_degl     : std_logic;
   signal sda_degl_d0  : std_logic;

   signal scl_degl     : std_logic;
   signal scl_degl_d0  : std_logic;

   signal state          : t_state;

   signal scl_rising         : std_logic;
   signal scl_falling        : std_logic;

begin

  cmp_scl_debouncer: i2c_debouncer
  generic map
  (
    g_LENGTH   => 6
  )
  port map
  (
    rst          => rst_i,
    clk          => clk_i,
    input        => scl_i,
    output       => scl_degl,
    glitch_mask  => c_GLITCH_MASK
  );

   
  cmp_sda_debounce: i2c_debouncer
  generic map
  (
    g_LENGTH   => 6
  )
  port map
  (
    rst          => rst_i,
    clk          => clk_i,
    input        => sda_i,
    output       => sda_degl,
    glitch_mask  => c_GLITCH_MASK
  );

   p_degl_delay: process(clk_i)
   begin
     if rising_edge(clk_i) then
       if (rst_i = '1') then
         sda_degl_d0 <= '0';
         scl_degl_d0 <= '0';
       else
         sda_degl_d0 <= sda_degl;
         scl_degl_d0 <= scl_degl;
       end if;
     end if;     
   end process p_degl_delay;

  p_fsm: process(clk_i)
  begin
    if rising_edge(clk_i) then
      if (rst_i = '1') then
        state   <= ST_IDLE;
        start_o <= '0';
        stop_o  <= '0';
        rcved_o <= '0';
        done_o  <= '0';
      else
        case state is
          when ST_IDLE =>
          when others =>
        end case;
      end if;
    end if;
  end process p_fsm;


--   -- This is the process that samples the scl for detecting
--   -- rise and falling edges
--  reg_proc: process (clk_i)
--  begin
--    if rising_edge(clk_i) then
--      if (rst_i = '1') then
--        scl_falling <= '0';
--        scl_rising  <= '0';
--      elsif (scl_degl = '0') and  (scl_degl_d0 = '1') then
--        scl_falling <= '1';
--      elsif (scl_degl = '1') and  (scl_degl_d0 = '0') then
--        scl_rising  <= '1';
--      else
--        scl_falling <= '0';
--        scl_rising  <= '0';
--      end if;
--    end if;
--  end process;


--  -- Combinatorial process to update the outputs.
--  p_comb_output: process(state)
--  begin
--    start_o     <= '0';
--    stop_o     <= '0';
--    rcved_o     <= '0';
--    done_o      <= '0';
--
--    case state is
--
--      when  R0_RESET        =>
--        null;
--
--      when  S0_IDLE         =>
--        null;
--
--      when  S1A_HIGH   =>
--        rcved_o     <= '1';
--        done_o      <= '1';
--
--      when  S1B_LOW    =>
--        rcved_o     <= '0';
--        done_o      <= '1';
--
--      when  S2A_START  =>
--        start_o     <= '1';
--        done_o      <= '1';
--
--      when  S2B_STOP_DETECT =>
--        stop_o     <= '1';
--        done_o      <= '1';
--
--      when  Q1_ERROR   =>
--        null;
--
--      when  others     =>
--        null;
--
--    end case;
--  end process p_comb_output;

--   -- The fsm of this module, later on the sda sampled line is
--   -- validated in the falling edge of scl.
--  p_fsm: process(clk_i)
--  begin
--    if rising_edge(clk_i) then
--      if (rst_i = '1') then
--        state   <= R0_RESET;
--      elsif scl_falling = '1' then
--      -- After a detection of a falling edge we update the
--      -- detection of a '0', a '1' and a start condition.
--        case state is
--          when S1A_HIGH_TMP  =>
--            state <= S1A_HIGH;
--
--          when S1B_LOW_TMP   =>
--            state <= S1B_LOW;
--          
--          when S2A_START_TMP =>
--            state <= S2A_START;
--          
--          when others =>
--            state <= S0_IDLE;
--        end case;
--
--      elsif (scl_rising = '1') then
--      -- When a rising edge is detected we annotate the first value
--      -- in SDA: either a temporary '0' or '1'
--        if (sda_degl_d0 = '1') then
--          state <= S1A_HIGH_TMP;
--        else
--          state <= S1B_LOW_TMP;
--        end if;
--
--      else
--        -- When we are in high level of a scl cycle, we keep on updating
--        -- the FSM
--        if (scl_degl = '1') then
--          case state is
--            -- Just for random bit swapped coverage.
--            when S0_IDLE =>
--              if (sda_degl = '1') then
--                state <= S1A_HIGH_TMP;
--              else
--                state <= S1B_LOW_TMP;
--              end if;
--
--            when S1A_HIGH_TMP  =>
--              if sda_degl = '0' then
--                -- The detection of the start condition will be reported
--                -- in the next SCL rising edge.
--                state <= S2A_START_TMP;
--              end if;
--
--            when S1B_LOW_TMP   =>
--              if sda_degl = '1' then
--                -- The detection of the pause condition MUST be
--                -- reported immediately.
--                state <= S2B_STOP_DETECT;
--              end if;
--
--            when S2A_START_TMP =>
--              if (sda_degl = '1') then
--                --! This happens if the deglitching is not enough
--                state <= Q1_ERROR;
--              end if;
--
--            when others =>
--              state <= S0_IDLE;
--          end case;
--        else
--          if (scl_degl_d0 = '0') then
--            state <= S0_IDLE;
--          end if;
--        end if;
--      end if;
--    end if;
--  end process p_fsm;

end behav;
