--==============================================================================
-- CERN (BE-CO-HT)
-- I2C slave core
--==============================================================================
--
-- author: Theodor Stana (t.stana@cern.ch)
--
-- date of creation: 2013-03-13
--
-- version: 1.0
--
-- description: 
--
-- dependencies:
--
-- references:
-- 
--==============================================================================
-- GNU LESSER GENERAL PUBLIC LICENSE
--==============================================================================
-- This source file is free software; you can redistribute it and/or modify it
-- under the terms of the GNU Lesser General Public License as published by the
-- Free Software Foundation; either version 2.1 of the License, or (at your
-- option) any later version. This source is distributed in the hope that it
-- will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- See the GNU Lesser General Public License for more details. You should have
-- received a copy of the GNU Lesser General Public License along with this
-- source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html
--==============================================================================
-- last changes:
--    2013-03-13   Theodor Stana     t.stana@cern.ch     File created
--==============================================================================
-- TODO: 
--    - Stop condition
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;


entity i2c_slave is
  port
  (
    -- Clock, reset ports
    clk_i      : in  std_logic;
    rst_n_i    : in  std_logic;

    -- I2C lines
    scl_i      : in  std_logic;
    scl_o      : out std_logic;
    scl_en_o   : out std_logic;
    sda_i      : in  std_logic;
    sda_o      : out std_logic;
    sda_en_o   : out std_logic;

    -- Slave address
    i2c_addr_i : in  std_logic_vector(6 downto 0);
    
    -- ACK input, should be set after done_o = '1'
    -- '0' - ACK
    -- '1' - NACK
    ack_n_i    : in  std_logic;

    -- I2C bus operation, set after address detection
    -- '0' - write
    -- '1' - read
    op_o       : out std_logic;

    -- Byte to send, should be loaded while done_o = '1'
    tx_byte_i  : in  std_logic_vector(7 downto 0);

    -- Received byte, valid after done_o = '1'
    rx_byte_o  : out std_logic_vector(7 downto 0);

    -- Done signal, valid when
    -- * received address matches i2c_addr_i, signaling valid op_o;
    -- * a byte was received, signaling valid rx_byte_o and an ACK/NACK should be
    -- sent to master;
    -- * sent a byte, should set tx_byte_i.
    done_o     : out std_logic
  );
end entity i2c_slave;


architecture behav of i2c_slave is

  --============================================================================
  -- Type declarations
  --============================================================================
  type t_state is 
    (
      ST_IDLE,
      ST_STA,
      ST_ADDR,
      ST_ADDR_ACK,
      ST_RD,
      ST_RD_WAIT_ACK,
      ST_RD_ACK,
      ST_WR_LOAD_TXSR,
      ST_WR,
      ST_WR_WAIT_ACK,
      ST_WR_ACK,
      ST_STO
    );

  --============================================================================
  -- Constant declarations
  --============================================================================

  --============================================================================
  -- Component declarations
  --============================================================================
  component glitch_filt is
    generic
    (
    -- Length of glitch filter:
    -- g_len = 1 => data width should be > 1 clk_i cycle
    -- g_len = 2 => data width should be > 2 clk_i cycle
    -- etc.
      g_len : natural := 4
    );
    port
    (
      clk_i   : in  std_logic;
      rst_n_i : in  std_logic;

    -- Data input
      dat_i   : in  std_logic;

    -- Data output
    -- latency: g_len+1 clk_i cycles
      dat_o   : out std_logic
    );
  end component glitch_filt;

  --============================================================================
  -- Signal declarations
  --============================================================================
  -- Deglitched signals and delays for SCL and SDA lines
  signal scl_degl, scl_degl_d0 : std_logic;
  signal sda_degl, sda_degl_d0 : std_logic;
  signal scl_falling           : std_logic;
  signal sda_falling           : std_logic;
  signal sda_rising            : std_logic;


  -- FSM
  signal state : t_state;

  -- RX and TX shift registers
  signal txsr : std_logic_vector(7 downto 0);
  signal rxsr : std_logic_vector(7 downto 0);

  -- Bit counter on RX & TX
  signal bit_cnt : unsigned(2 downto 0);

--==============================================================================
--  architecture begin
--==============================================================================
begin

  --============================================================================
  -- I/O logic
  --============================================================================
  -- No clock stretching implemented, always disable SCL line
  scl_o     <= '0';
  scl_en_o  <= '0';
  
  -- Assign RX byte output
  rx_byte_o <= rxsr;

  --============================================================================
  -- Deglitching logic
  --============================================================================
  -- Generate deglitched SCL signal with 54-ns max. glitch width
  cmp_scl_deglitch: glitch_filt
    generic map
    (
      g_len => 7
    )
    port map
    (
      clk_i   => clk_i,
      rst_n_i => rst_n_i,
      dat_i   => scl_i,
      dat_o   => scl_degl
    );

  -- and create a delayed version of this signal, together with one-tick-long
  -- falling-edge detection signal
  p_scl_degl_d0: process(clk_i) is
  begin
    if rising_edge(clk_i) then
      if (rst_n_i = '0') then
        scl_degl_d0  <= '0';
        scl_falling <= '0';
      else
        scl_degl_d0  <= scl_degl;
        scl_falling <= '0';
        if (scl_degl = '0') and (scl_degl_d0 = '1') then
          scl_falling <= '1';
        end if;
      end if;
    end if;
  end process p_scl_degl_d0;
  
  -- Generate deglitched SDA signal with 54-ns max. glitch width
  cmp_sda_deglitch: glitch_filt
    generic map
    (
      g_len => 7
    )
    port map
    (
      clk_i   => clk_i,
      rst_n_i => rst_n_i,
      dat_i   => sda_i,
      dat_o   => sda_degl
    );

  -- and create a delayed version of this signal, together with one-tick-long
  -- falling- and rising-edge detection signals
  p_sda_degl_d0: process(clk_i) is
  begin
    if rising_edge(clk_i) then
      if (rst_n_i = '0') then
        sda_degl_d0 <= '0';
        sda_rising  <= '0';
        sda_falling <= '0';
      else
        sda_degl_d0 <= sda_degl;
        
        sda_rising  <= '0';
        if (sda_degl = '1') and (sda_degl_d0 = '0') then
          sda_rising <= '1';
        end if;
        
        sda_falling <= '0';
        if (sda_degl = '0') and (sda_degl_d0 = '1') then
          sda_falling <= '1';
        end if;
      end if;
    end if;
  end process p_sda_degl_d0;

  --============================================================================
  -- FSM logic
  --============================================================================
  p_fsm: process (clk_i) is
  begin
    if rising_edge(clk_i) then
      if (rst_n_i = '0') then
        state    <= ST_IDLE;
        bit_cnt  <= (others => '0');
        rxsr     <= (others => '0');
        txsr     <= (others => '0');
        sda_o    <= '0';
        sda_en_o <= '0';
        done_o   <= '0';
        op_o     <= '0';
      elsif (sda_falling = '1') and (scl_degl = '1') then
        state <= ST_STA;
      elsif (sda_rising = '1') and (scl_degl = '1') then
        state <= ST_IDLE;
      else
        case state is
          when ST_IDLE =>
            bit_cnt  <= (others => '0');
            sda_o    <= '0';
            sda_en_o <= '0';
            done_o   <= '0';
            -- if (sda_falling = '1') and (scl_degl = '1') then
            --   state  <= ST_STA;
            -- end if;

          when ST_STA =>
            bit_cnt <= (others => '0');
            if (scl_falling = '1') then
              state <= ST_ADDR;
            end if;

          when ST_ADDR =>
            if (scl_falling = '1') then
              rxsr    <= rxsr(6 downto 0) & sda_i;
              bit_cnt <= bit_cnt + 1;
              if (bit_cnt = 7) then
                state <= ST_ADDR_ACK;
              end if;
            end if;

          when ST_ADDR_ACK =>
            sda_en_o <= '1';
            -- if rcvd address is ours, ACK, check R/W bit and 
            -- go to appropriate state
            if (rxsr(7 downto 1) = i2c_addr_i) then
              sda_o  <= '0';
              op_o   <= rxsr(0);
              done_o <= '1';
              if (scl_falling = '1') then
                sda_en_o <= '0';
                if (rxsr(0) = '0') then
                  state <= ST_RD;
                else
                  state <= ST_WR_LOAD_TXSR;
                end if;
              end if;
            -- if received address is not ours, NACK and go back to IDLE
            else
              sda_o <= '1';
              state <= ST_IDLE;
            end if;

          when ST_RD =>
            done_o     <= '0';
            if (scl_falling = '1') then
              rxsr     <= rxsr(6 downto 0) & sda_i;
              bit_cnt  <= bit_cnt + 1;
              if (bit_cnt = 7) then
                state  <= ST_RD_WAIT_ACK;
                done_o <= '1';
              end if;
            end if;

          when ST_RD_WAIT_ACK =>
            bit_cnt <= bit_cnt + 1;
            if (bit_cnt = 7) then
              state <= ST_RD_ACK;
            end if;

          when ST_RD_ACK =>
            sda_en_o <= '1';
            sda_o    <= ack_n_i;
            if (scl_falling = '1') then
              sda_en_o <= '0';
              if (ack_n_i = '1') then
                state  <= ST_RD;
                -- done_o <= '0';
              else
                state  <= ST_IDLE;
              end if;
            end if;

          when ST_WR_LOAD_TXSR =>
            txsr  <= tx_byte_i;
            state <= ST_WR;

          when ST_WR =>
            done_o   <= '0';
            sda_en_o <= '1';
            sda_o    <= txsr(7);
            if (scl_falling = '1') then
              txsr    <= txsr(6 downto 0) & '0';
              bit_cnt <= bit_cnt + 1;
              if (bit_cnt = 7) then
                sda_en_o <= '0';
                state    <= ST_WR_WAIT_ACK;
                done_o   <= '1';
              end if;
            end if;

          when ST_WR_WAIT_ACK =>
            bit_cnt <= bit_cnt + 1;
            if (bit_cnt = 7) then
              state <= ST_WR_ACK;
            end if;

          when ST_WR_ACK =>
            if (scl_falling = '1') then
              if (sda_degl = '0') then
                state  <= ST_WR_LOAD_TXSR;
                -- done_o <= '0';
              else
                state  <= ST_IDLE;
              end if;
            end if;

          -- when ST_STO =>
          --   if (sda_rising = '1') and (scl_degl = '1') then
          --     state <= ST_IDLE;
          --   end if;

          when others =>
            state <= ST_IDLE;
        end case;
      end if;
    end if;
  end process p_fsm;

end architecture behav;
--==============================================================================
--  architecture end
--==============================================================================
