--------------------------------------------------------------------------------
-- Company:        CERN, BE-CO-HT CHANGED
-- Engineer:       Carlos Gil Soriano
--
-- Create Date:    18:15:56 11/09/2011
-- Design Name:    A I2C slave with wishbone slave output and interrupt for MCU
-- Module Name:    vme64x_i2c - behav
-- Project Name:   CTDAH
-- Target Devices:
-- Tool versions:
-- Description:
--
-- Dependencies:
--
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
--
----------------------------------------------------------------------------------
library IEEE;
library work;

use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;

entity vme64x_i2c is
  port
  (
    -- Clock, reset
    clk_i           : in  std_logic;
    rst_n_i         : in  std_logic;

    -- I2C lines
    sda_en_o        : out std_logic;
    sda_i           : in  std_logic;
    sda_o           : out std_logic;
    scl_en_o        : out std_logic;
    scl_i           : in  std_logic;
    scl_o           : out std_logic;

    -- I2C address and status
    i2c_addr_i      : in  std_logic_vector(6 downto 0);
    i2c_done_o      : out std_logic;

    -- Wishbone master signals
    wb_master_stb_o : out std_logic;
    wb_master_cyc_o : out std_logic;
    wb_master_sel_o : out std_logic_vector(3 downto 0);
    wb_master_we_o  : out std_logic;
    wb_master_dat_i : in  std_logic_vector(31 downto 0);
    wb_master_dat_o : out std_logic_vector(31 downto 0);
    wb_master_adr_o : out std_logic_vector(31 downto 0);
    wb_master_ack_i : in  std_logic;
    wb_master_rty_i : in  std_logic;
    wb_master_err_i : in  std_logic
  );
end vme64x_i2c;

architecture behav of vme64x_i2c is

  --============================================================================
  -- Component declarations
  --============================================================================
  -- I2C slave component
  -- (use: provide received byte, stores byte to send)
  component i2c_slave is
    port
    (
    -- Clock, reset ports
      clk_i      : in  std_logic;
      rst_n_i    : in  std_logic;

    -- I2C lines
      scl_i      : in  std_logic;
      scl_o      : out std_logic;
      scl_en_o   : out std_logic;
      sda_i      : in  std_logic;
      sda_o      : out std_logic;
      sda_en_o   : out std_logic;

    -- Slave address
      i2c_addr_i : in  std_logic_vector(6 downto 0);

    -- ACK input, should be set after done_o = '1'
    -- '0' - ACK
    -- '1' - NACK
      ack_n_i    : in  std_logic;

    -- I2C bus operation, set after address detection
    -- '0' - write
    -- '1' - read
      op_o       : out std_logic;

    -- Byte to send, should be loaded while done_o = '1'
      tx_byte_i  : in  std_logic_vector(7 downto 0);

    -- Received byte, valid after done_o = '1'
      rx_byte_o  : out std_logic_vector(7 downto 0);

    -- Done signal, valid when
    -- * received address matches i2c_addr_i, signaling valid op_o;
    -- * a byte was received, signaling valid rx_byte_o and an ACK/NACK should be
    -- sent to master;
    -- * sent a byte, should set tx_byte_i.
      done_o     : out std_logic
    );
  end component i2c_slave;
  
  --============================================================================
  -- Signal declarations
  --============================================================================
  -- Slave component signals
  signal ack_n   : std_logic;
  signal op      : std_logic;
  signal tx_byte : std_logic_vector(7 downto 0);
  signal rx_byte : std_logic_vector(7 downto 0);
  signal done    : std_logic;

begin

  --============================================================================
  -- Slave component instantiation and connection
  --============================================================================
  cmp_i2c_slave: i2c_slave
    port map
    (
      clk_i      => clk_i,
      rst_n_i    => rst_n_i,

      -- I2C lines
      scl_i      => scl_i,
      scl_o      => scl_o,
      scl_en_o   => scl_en_o,
      sda_i      => sda_i,
      sda_o      => sda_o,
      sda_en_o   => sda_en_o,

      -- Slave address
      i2c_addr_i => i2c_addr_i,

      -- ACK input, should be set after done_o = '1'
      -- '0' - ACK
      -- '1' - NACK
      ack_n_i    => ack_n,

      -- I2C bus operation, set after address detection
      -- '0' - write
      -- '1' - read
      op_o       => op,

      -- Byte to send, should be loaded while done_o = '1'
      tx_byte_i  => tx_byte,

      -- Received byte, valid after done_o = '1'
      rx_byte_o  => rx_byte,

      -- Done signal, valid when
      -- * received address matches i2c_addr_i, signaling valid op_o;
      -- * a byte was received, signaling valid rx_byte_o and an ACK/NACK should be
      -- sent to master;
      -- * sent a byte, should set tx_byte_i.
      done_o     => done
    );

  
end behav;
